==> pan-0.7.0.tar.gz: pan-0.7.0/pan/queue-item-headers.c <==
/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <gnome.h>

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include <glib.h>

#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article.h"
#include "article-db.h"
#include "date.h"
#include "debug.h"
#include "log.h"
#include "queue-item-headers.h"

#include "articlelist.h"
#include "grouplist.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static void queue_item_headers_class_init (QueueItemHeadersClass *klass);

static void queue_item_headers_init (QueueItemHeaders *item);

static gint queue_item_headers_run (QueueItem* item);

static void queue_item_headers_destroy (GtkObject* item);

static gchar* queue_item_headers_describe (StatusItem *item);

static GSList * nntp_articlelist_download (GtkObject* item, int *count);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

extern gboolean must_exit;

/***********
************  Public
***********/

/***********
************  Private
***********/

static QueueItemClass *parent_class = NULL;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

GtkType
queue_item_headers_get_type (void)
{
	static GtkType queue_item_headers_type = 0;

	if (!queue_item_headers_type)
	{
		static const GtkTypeInfo queue_item_headers_info =
		{
			"QueueItemHeaders",
			sizeof(QueueItemHeaders),
			sizeof(QueueItemHeadersClass),
			(GtkClassInitFunc) queue_item_headers_class_init,
			(GtkObjectInitFunc) queue_item_headers_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};

		queue_item_headers_type = gtk_type_unique (
			TYPE_QUEUE_ITEM,
			&queue_item_headers_info);
	}

	return queue_item_headers_type;
}

static void
queue_item_headers_class_init (QueueItemHeadersClass *klass)
{
        parent_class = gtk_type_class (TYPE_QUEUE_ITEM);

	GTK_OBJECT_CLASS(klass)->destroy = queue_item_headers_destroy;
	STATUS_ITEM_CLASS(klass)->describe = queue_item_headers_describe;
	QUEUE_ITEM_CLASS(klass)->run = queue_item_headers_run;
}

static void
queue_item_headers_init (QueueItemHeaders *item)
{
	item->gdata = NULL;
	item->download_type = HEADERS_ALL;
}

/*****
******
*****/

GtkObject*
queue_item_headers_new (server_data *sdata,
	                gboolean high_priority,
			group_data *gdata,
			HeaderDownloadType dl_type)
{
	QueueItemHeaders *item = NULL;
	gboolean need_socket = dl_type != HEADERS_LOCAL;

	/* sanity checks */
	g_return_val_if_fail(sdata!=NULL, NULL);
	g_return_val_if_fail(gdata!=NULL, NULL);
	item = QUEUE_ITEM_HEADERS(gtk_type_new(TYPE_QUEUE_ITEM_HEADERS));

	/* construct superclass... */
	queue_item_construct (QUEUE_ITEM(item),
		sdata, high_priority, need_socket);

	/* construct this class... */
	item->gdata = gdata;
	item->download_type = dl_type;

	return GTK_OBJECT(item);
}


/*****
******
*****/

static gchar*
queue_item_headers_describe (StatusItem* item)
{
	const gchar* action = NULL;

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);
	g_return_val_if_fail (IS_QUEUE_ITEM_HEADERS(item), NULL);
	g_return_val_if_fail (GTK_OBJECT_CONSTRUCTED(item), NULL);

	switch (QUEUE_ITEM_HEADERS(item)->download_type) {
		case HEADERS_LOCAL: action=_("cached headers"); break;
		case HEADERS_ALL: action=_("all headers"); break;
		case HEADERS_NEW: action=_("new headers"); break;
		case HEADERS_SAMPLE: action=_("a sample of headers"); break;
		default: g_assert_not_reached();
	}

	return g_strdup_printf ( _("Getting %s for %s"),
		action,
		QUEUE_ITEM_HEADERS(item)->gdata->name );
}

/*****
******
*****/

static gint
queue_item_headers_run (QueueItem* item)
{
	QueueItemHeaders *item_h = QUEUE_ITEM_HEADERS(item);
	int count = 0;
	GSList *alist = NULL;
	article_db db = NULL;
	group_data *gdata = NULL;
	server_data *sdata = NULL;
	gboolean active;
	int total = 0;

	if (item_h->download_type == HEADERS_LOCAL) {
		articlelist_load (
			item->sdata, item_h->gdata, STATUS_ITEM(item));
		return 0;
	}

	sdata = item->sdata;
	gdata = item_h->gdata;
	total = group_get_attrib_i (sdata, gdata,"Total");

	/* get the article headers... */
	alist = nntp_articlelist_download (GTK_OBJECT(item), &count);

	if ( !alist ) {
		status_item_emit_error (STATUS_ITEM(item),
			_("No articles retrieved for group %s"),
			gdata->name);
		return -1;
	}

	db = ahdb_ref ( sdata, gdata );

	log_add_va (
		_("Got %d article headers from %s"),
		count,
		gdata->name);

	active = sdata==articlelist_get_current_server() &&
	         gdata==articlelist_get_current_group();

	if (item_h->download_type == HEADERS_NEW)
	{
		GSList* old_headers = NULL;

		status_item_emit_init_steps (STATUS_ITEM(item), 	
			(active ? total : 0) /* Load of old headers */
			+ count /* Save of new headers */
			+ (active?(total+count)*3:0)); /* alist_set_contents */
		
		/* get old headers, if necessary */
		if (active)
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Loading Old Headers"));
			old_headers = ahdb_get_all (
				db, status_item_next_step_gfunc, item);
			if (g_slist_length(old_headers) != total)
				g_warning ("Expected %d articles; loaded %d",
					total, g_slist_length(old_headers));
		}

		/* save the new headers... */
		status_item_emit_status (STATUS_ITEM(item),
			_("Saving New Headers"));
		ahdb_save_all (db, alist, status_item_next_step_gfunc, item);

		/* update the group properties... */
		if ( active )
		{
			/* join new headers with any old ones.
			   note that since old ones were loaded from disk,
			   their "new" flag is turned off automatically */
			alist = g_slist_concat (alist, old_headers);
			group_set_attrib_i(
				sdata, gdata, "Total", g_slist_length(alist));
		}
		else
		{
			group_add_attrib_i (
				sdata, gdata, "Total", count);
		}
	}
	else /* download all -or- headers */
	{
		status_item_emit_init_steps (
			STATUS_ITEM(item),
			count /* Save of new headers */
			+ count /* Killfile */
			+ count /* Thread */	
			+ count); /* Display */
		
		/* Save all the fresh new headers */
		ahdb_erase_all (db);
		group_set_attrib_i (sdata, gdata, "Read", 0);
		group_set_attrib_i (sdata, gdata, "Total", count);
		status_item_emit_status (STATUS_ITEM(item),
				_("Saving Headers"));
		ahdb_save_all (db, alist, status_item_next_step_gfunc, item);
	}

	/* update grouplist */
	grouplist_update_row (gdata);

	/* take care of the alist */
	if (active)
	{
		const int total = group_get_attrib_i (sdata, gdata, "Total");
		g_assert (total == g_slist_length (alist));
		articlelist_set_contents (
			sdata,  gdata, alist, total, STATUS_ITEM(item));
	}
	else
	{
		g_slist_foreach ( alist, article_free_gfunc, NULL );
		g_slist_free ( alist );
	}

	status_item_emit_progress (STATUS_ITEM(item), 0);

	/* cleanup */	
	ahdb_sync ( db );
	ahdb_unref ( db );

	return 0;
}

/*****
******
*****/

static void
queue_item_headers_destroy (GtkObject* object)
{
        debug (DEBUG_QUEUE,
		"GtkObject '%p' in queue_item_headers_destroy",object);

	/* call superclass */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		(*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/*****
******
*****/

static GSList *
nntp_articlelist_download (GtkObject* item,
			   int *count)
{
	PanSocket *sock = QUEUE_ITEM(item)->sock;
	server_data *sdata = QUEUE_ITEM(item)->sdata;
	group_data *gdata = QUEUE_ITEM_HEADERS(item)->gdata;
	HeaderDownloadType dl_type = QUEUE_ITEM_HEADERS(item)->download_type;

	const char *buffer = NULL;
	article_data *adata = NULL;
	gint total, total_new;
	int steps=0;
	int first=0, last=0, p_last=0;
	const time_t current_time=time(0);
	GSList *alist = NULL;
	gboolean read_status = 0;
	const int expire_seconds = (60*60*24) *
		gnome_config_get_int_with_default ("/Pan/Articles/Expiration=30", NULL);

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, NULL);
	g_return_val_if_fail (sdata!=NULL, NULL);
	g_return_val_if_fail (gdata!=NULL, NULL);
	g_return_val_if_fail (count!=NULL, NULL);

	/* change to the right group */
	if (pan_socket_putline_va (sock, "GROUP %s\r\n", gdata->name)) {
		status_item_emit_sockwrite_err (STATUS_ITEM(item));
		return NULL;
	}
	if (pan_socket_getline (sock, &buffer)) {
		status_item_emit_sockread_err (STATUS_ITEM(item));
		return NULL;
	}
	if (strtol(buffer, NULL, 10) != 211) {
		status_item_emit_error (STATUS_ITEM(item),
			_("(Couldn't select group '%s'"), gdata->name);
		return NULL;
	}
	sscanf (buffer, "%*d %d %d %d", &total, &first, &last);

	p_last = group_get_attrib_i (sdata, gdata, "Last");

	total_new = last - p_last;

	if (total == '0') {
		status_item_emit_status (STATUS_ITEM(item),
			_("No articles found for group '%s'"), gdata->name);
		return NULL;
	}

	if ((dl_type == HEADERS_NEW) && p_last)
	{
		if (p_last == last)
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("No new articles found for group '%s'"),
				gdata->name);
			return NULL;
		}
		else
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting new article headers..."));

			if (pan_socket_putline_va (sock,
					"XOVER %d-%d\r\n", p_last, last))
			{
				status_item_emit_sockwrite_err(
					STATUS_ITEM(item));
				return NULL;
			}
		}
	}
	else
	{
		if ((total > 150) && ((dl_type==HEADERS_SAMPLE)))
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting sample"));
			last = first + 150;
		}
		else
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting all article headers"));
		}

		if (pan_socket_putline_va (sock,"XOVER %d-%d\r\n",first,last))
		{
			status_item_emit_sockwrite_err (STATUS_ITEM(item));
			return NULL;
		}
	}

	if (pan_socket_getline (sock, &buffer)) {
		status_item_emit_sockread_err (STATUS_ITEM(item));
		return NULL;
	}

	if (strtol(buffer, (char **) NULL, 10) != 224) {
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't retrieve headers\nServer said: %s"), buffer);
		return NULL;
	}

	steps = ((dl_type==HEADERS_NEW) && (group_get_attrib_i (sdata, gdata,"Total")>0))
		? total_new
		: total;
	status_item_emit_init_steps (STATUS_ITEM(item), steps);

	while (!must_exit && !((read_status = pan_socket_getline (sock, &buffer))))
	{
		int i = 0;
		int line_qty = 0;
		int part = 0;
	        int parts = 0;
		time_t date = 0;
		gchar **sd = NULL;
		gchar *s = NULL;

		/* handle end-of-list */
		if (!strncmp(buffer, ".\r\n", 3))
			break;

		/* split the header */
		sd = g_strsplit (buffer, "\t", 8);

		/* check for incomplete header */
		line_qty = 0;
		while ( sd[line_qty]!=NULL )
			++line_qty;
		if (line_qty<8) {
			status_item_emit_error (STATUS_ITEM(item),
				_("Corrupt header ``%s'' skipped"),buffer);
			g_strfreev (sd);
			continue;
		}

		/* check for expired message */
		date = parse_date (sd[3]);
		if (expire_seconds < current_time-date) {
			g_strfreev (sd);
			continue;
		}

		/* create the article data */
		adata = g_new0 ( article_data, 1 );
		adata->number = sd[0];
		adata->subject = sd[1];
		adata->author = sd[2];
		adata->date = date;
		adata->message_id = sd[4];
		adata->references = sd[5];
		adata->linecount = strtol(sd[7], NULL, 10);
		adata->state = STATE_NEW;

		/* cleanup */
		g_free (sd[3]);
		g_free (sd[6]);
		g_free (sd[7]);
		g_free (sd[8]); /* NULL or server-optional extra headers */
		g_free (sd);

		/* Look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = adata->subject;
		for (i = strlen(s) - 1; i > 0; --i) {
			if ((s[i] == ')') || (s[i] == ']')) {
				for (--i; isdigit((int)s[i]); --i);
				if (s[i] == '/') {
					parts = atoi(s + i + 1);
					for (--i; isdigit((int)s[i]); --i);
					if ((s[i] == '(') || (s[i] == '[')) {
						part = atoi(s + i + 1);
						break;
					}
				}
			}
		}

		/* Verify Multipart info */
		if ((parts > 1) && (part <= parts)) {
			/* We found one! */
			adata->parts = parts;
			adata->part = part;
		}
		else {
			adata->parts = 0;
			adata->part = 0;
		}

		/* Add the article to the article list */
		++*count;
		alist = g_slist_prepend (alist, adata);

		/* update the count & progress ui */
		status_item_emit_next_step (STATUS_ITEM(item));
		if ( *count % 29 == 0 ) {
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting new article headers %d of %d"),
				*count, steps );
		}
	}
	if (read_status!=0 || must_exit)
	{
		if (read_status!=0)
			status_item_emit_sockread_err (STATUS_ITEM(item));
		g_slist_foreach (alist, article_free_gfunc, NULL);
		g_slist_free (alist);
		return NULL;
	}
	
	group_set_attrib_i (sdata, gdata, "First", first);
	group_set_attrib_i (sdata, gdata, "Last", last);

	return alist;
}

==> pan-0.7.3.tar.gz: pan-0.7.3/pan/queue-item-headers.c <==
/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <gnome.h>

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include <glib.h>

#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article.h"
#include "article-db.h"
#include "date.h"
#include "debug.h"
#include "log.h"
#include "queue-item-headers.h"

#include "articlelist.h"
#include "grouplist.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint queue_item_headers_run (QueueItem* item);

static gchar* queue_item_headers_describe (const StatusItem *item);

static GSList* nntp_articlelist_download (QueueItemHeaders* item, int *count);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

/*****
******
*****/

PanObject*
queue_item_headers_new (server_data *sdata,
	                gboolean high_priority,
			group_data *gdata,
			HeaderDownloadType dl_type)
{
	QueueItemHeaders *item = g_new0 (QueueItemHeaders, 1);
	gboolean need_socket = dl_type != HEADERS_LOCAL;

	debug (DEBUG_PAN_OBJECT, "queue_item_headers_new: %p", item);

	queue_item_constructor (QUEUE_ITEM(item),
		queue_item_destructor,
		queue_item_headers_describe,
		queue_item_headers_run, sdata, high_priority, need_socket);

	item->gdata = gdata;
	item->download_type = dl_type;

	return PAN_OBJECT(item);
}


/*****
******
*****/

static gint
queue_item_headers_run (QueueItem* item)
{
	QueueItemHeaders *item_h = QUEUE_ITEM_HEADERS(item);
	int count = 0;
	GSList *alist = NULL;
	article_db db = NULL;
	group_data *gdata = NULL;
	server_data *sdata = NULL;
	gboolean active;
	int total = 0;

	if (item_h->download_type == HEADERS_LOCAL) {
		articlelist_load (
			item->sdata, item_h->gdata, STATUS_ITEM(item));
		return 0;
	}

	sdata = item->sdata;
	gdata = item_h->gdata;
	total = group_get_attrib_i (sdata, gdata,"Total");

	/* get the article headers... */
	alist = nntp_articlelist_download (QUEUE_ITEM_HEADERS(item), &count);

	if ( !alist ) {
		status_item_emit_error (STATUS_ITEM(item),
			_("No articles retrieved for group %s"),
			gdata->name);
		return -1;
	}

	db = ahdb_ref ( sdata, gdata );

	log_add_va (
		_("Downloaded %d article headers from %s"),
		count,
		gdata->name);

	active = sdata==articlelist_get_current_server() &&
	         gdata==articlelist_get_current_group();

	if (item_h->download_type == HEADERS_NEW)
	{
		GSList* old_headers = NULL;

		status_item_emit_init_steps (STATUS_ITEM(item), 	
			(active ? total : 0) /* Load of old headers */
			+ count /* Save of new headers */
			+ (active?(total+count)*3:0)); /* alist_set_contents */
		
		/* get old headers, if necessary */
		if (active)
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Loading Old Headers"));
			old_headers = ahdb_get_all (
				db, status_item_next_step_gfunc, item);
			if (g_slist_length(old_headers) != total)
				g_warning ("Expected %d articles; loaded %d",
					total, g_slist_length(old_headers));
		}

		/* save the new headers... */
		status_item_emit_status (STATUS_ITEM(item),
			_("Saving New Headers"));
		ahdb_save_all (db, alist, status_item_next_step_gfunc, item);

		/* update the group properties... */
		if ( active )
		{
			/* join new headers with any old ones.
			   note that since old ones were loaded from disk,
			   their "new" flag is turned off automatically */
			alist = g_slist_concat (alist, old_headers);
			group_set_attrib_i(
				sdata, gdata, "Total", g_slist_length(alist));
		}
		else
		{
			group_add_attrib_i (
				sdata, gdata, "Total", count);
		}
	}
	else /* download all -or- headers */
	{
		status_item_emit_init_steps (
			STATUS_ITEM(item),
			count /* Save of new headers */
			+ count /* Killfile */
			+ count /* Thread */	
			+ count); /* Display */
		
		/* Save all the fresh new headers */
		ahdb_erase_all (db);
		group_set_attrib_i (sdata, gdata, "Read", 0);
		group_set_attrib_i (sdata, gdata, "Total", count);
		status_item_emit_status (STATUS_ITEM(item),
				_("Saving Headers"));
		ahdb_save_all (db, alist, status_item_next_step_gfunc, item);
	}

	/* update grouplist */
	grouplist_update_row (gdata);

	/* take care of the alist */
	if (active)
	{
		const int total = group_get_attrib_i (sdata, gdata, "Total");
		g_assert (total == g_slist_length (alist));
		articlelist_set_contents (
			sdata,  gdata, alist, total, STATUS_ITEM(item));
	}
	else
	{
		g_slist_foreach ( alist, article_free_gfunc, NULL );
		g_slist_free ( alist );
	}

	status_item_emit_progress (STATUS_ITEM(item), 0);

	/* cleanup */	
	ahdb_sync ( db );
	ahdb_unref ( db );

	return 0;
}

/*****
******
*****/

static GSList *
nntp_articlelist_download (QueueItemHeaders* item,
			   int *count)
{
	PanSocket *sock = QUEUE_ITEM(item)->sock;
	server_data *sdata = QUEUE_ITEM(item)->sdata;
	group_data *gdata = QUEUE_ITEM_HEADERS(item)->gdata;
	HeaderDownloadType dl_type = QUEUE_ITEM_HEADERS(item)->download_type;

	const char *buffer = NULL;
	article_data *adata = NULL;
	gint total, total_new;
	int steps=0;
	int first=0, last=0, p_last=0;
	const time_t current_time=time(0);
	GSList *alist = NULL;
	gboolean read_status = 0;
	const int expire_seconds = (60*60*24) *
		gnome_config_get_int_with_default ("/Pan/Articles/Expiration=30", NULL);

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, NULL);
	g_return_val_if_fail (sdata!=NULL, NULL);
	g_return_val_if_fail (gdata!=NULL, NULL);
	g_return_val_if_fail (count!=NULL, NULL);

	/* change to the right group */
	if (pan_socket_putline_va (sock, "GROUP %s\r\n", gdata->name)) {
		queue_item_emit_sockwrite_err (QUEUE_ITEM(item));
		return NULL;
	}
	if (pan_socket_getline (sock, &buffer)) {
		queue_item_emit_sockread_err (QUEUE_ITEM(item));
		return NULL;
	}
	if (strtol(buffer, NULL, 10) != 211) {
		status_item_emit_error (STATUS_ITEM(item),
			_("(Couldn't select group '%s'"), gdata->name);
		return NULL;
	}
	sscanf (buffer, "%*d %d %d %d", &total, &first, &last);

	p_last = group_get_attrib_i (sdata, gdata, "Last");

	total_new = last - p_last;

	if (total == '0') {
		status_item_emit_status (STATUS_ITEM(item),
			_("No articles found for group '%s'"), gdata->name);
		return NULL;
	}

	if ((dl_type == HEADERS_NEW) && p_last)
	{
		if (p_last == last)
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("No new articles found for group '%s'"),
				gdata->name);
			return NULL;
		}
		else
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting new article headers..."));

			if (pan_socket_putline_va (sock,
					"XOVER %d-%d\r\n", p_last, last))
			{
				queue_item_emit_sockwrite_err(QUEUE_ITEM(item));
				return NULL;
			}
		}
	}
	else
	{
		if ((total > 150) && ((dl_type==HEADERS_SAMPLE)))
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting sample"));
			last = first + 150;
		}
		else
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting all article headers"));
		}

		if (pan_socket_putline_va (sock,"XOVER %d-%d\r\n",first,last))
		{
			queue_item_emit_sockwrite_err (QUEUE_ITEM(item));
			return NULL;
		}
	}

	status_item_emit_status (STATUS_ITEM(item),
		_("Waiting for Response from Server..."));

	if (pan_socket_getline (sock, &buffer)) {
		queue_item_emit_sockread_err (QUEUE_ITEM(item));
		return NULL;
	}

	if (strtol(buffer, (char **) NULL, 10) != 224) {
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't retrieve headers\nServer said: %s"), buffer);
		return NULL;
	}

	steps = ((dl_type==HEADERS_NEW) && (group_get_attrib_i (sdata, gdata,"Total")>0))
		? total_new
		: total;
	status_item_emit_init_steps (STATUS_ITEM(item), steps);

	while (!QUEUE_ITEM(item)->abort && !((read_status = pan_socket_getline (sock, &buffer))))
	{
		int i = 0;
		int line_qty = 0;
		int part = 0;
	        int parts = 0;
		time_t date = 0;
		gchar **sd = NULL;
		gchar *s = NULL;

		/* handle end-of-list */
		if (!strncmp(buffer, ".\r\n", 3))
			break;

		/* split the header */
		sd = g_strsplit (buffer, "\t", 8);

		/* check for incomplete header */
		line_qty = 0;
		while ( sd[line_qty]!=NULL )
			++line_qty;
		if (line_qty<8) {
			status_item_emit_error (STATUS_ITEM(item),
				_("Corrupt header ``%s'' skipped"),buffer);
			g_strfreev (sd);
			continue;
		}

		/* check for expired message */
		date = parse_date (sd[3]);
		if (expire_seconds < current_time-date) {
			g_strfreev (sd);
			continue;
		}

		/* create the article data */
		adata = g_new0 ( article_data, 1 );
		adata->number = sd[0];
		adata->subject = sd[1];
		adata->author = sd[2];
		adata->date = date;
		adata->message_id = sd[4];
		adata->references = sd[5];
		adata->linecount = strtol(sd[7], NULL, 10);
		adata->state = STATE_NEW;

		/* cleanup */
		g_free (sd[3]);
		g_free (sd[6]);
		g_free (sd[7]);
		g_free (sd[8]); /* NULL or server-optional extra headers */
		g_free (sd);

		/* Look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = adata->subject;
		for (i = strlen(s) - 1; i > 0; --i) {
			if ((s[i] == ')') || (s[i] == ']')) {
				for (--i; isdigit((int)s[i]); --i);
				if (s[i] == '/') {
					parts = atoi(s + i + 1);
					for (--i; isdigit((int)s[i]); --i);
					if ((s[i] == '(') || (s[i] == '[')) {
						part = atoi(s + i + 1);
						break;
					}
				}
			}
		}

		/* Verify Multipart info */
		if ((parts > 1) && (part <= parts)) {
			/* We found one! */
			adata->parts = parts;
			adata->part = part;
		}
		else {
			adata->parts = 0;
			adata->part = 0;
		}

		/* Add the article to the article list */
		++*count;
		alist = g_slist_prepend (alist, adata);

		/* update the count & progress ui */
		status_item_emit_next_step (STATUS_ITEM(item));
		if ( *count % 29 == 0 ) {
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting new article headers %d of %d"),
				*count, steps );
		}
	}
	if (read_status!=0 || QUEUE_ITEM(item)->abort)
	{
		if (read_status!=0)
			queue_item_emit_sockread_err (QUEUE_ITEM(item));
		g_slist_foreach (alist, article_free_gfunc, NULL);
		g_slist_free (alist);
		return NULL;
	}
	
	group_set_attrib_i (sdata, gdata, "First", first);
	group_set_attrib_i (sdata, gdata, "Last", last);

	return alist;
}

static gchar*
queue_item_headers_describe (const StatusItem* item)
{
	const gchar* action = NULL;

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);

	switch (QUEUE_ITEM_HEADERS(item)->download_type)
	{
		case HEADERS_LOCAL: action=_("cached headers"); break;
		case HEADERS_ALL: action=_("all headers"); break;
		case HEADERS_NEW: action=_("new headers"); break;
		case HEADERS_SAMPLE: action=_("a sample of headers"); break;
		default: pan_warn_if_reached(); action=_("BUG IN CODE"); break;
	}

	return g_strdup_printf ( _("Getting %s for newsgroup `%s'"),
		action,
		QUEUE_ITEM_HEADERS(item)->gdata->name );
}
